#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import re
import logging
import subprocess
from pyradium.Tools import HashTools
from .BaseRenderer import BaseRenderer

_log = logging.getLogger(__spec__.name)

@BaseRenderer.register
class PlotRenderer(BaseRenderer):
	_NAME = "plot"
	_SET_TERMINAL_RE = re.compile(r"^\s*set\s+terminal.*$", flags = re.MULTILINE)

	@property
	def properties(self):
		return {
			"version":			1,
		}

	def _render_plot_png(self, src, width, height):
		with open(src) as f:
			source = f.read()
		source = self._SET_TERMINAL_RE.sub("", source)
		source = "set terminal pngcairo size %d,%d enhanced font 'Latin Modern Sans,24'\n" % (width, height) + source
		png_data = subprocess.check_output([ "gnuplot" ], input = source.encode("utf-8"), stderr = _log.subproc_target)
		return png_data

	def rendering_key(self, property_dict):
		src = property_dict["src"]
		return {
			"srchash":		HashTools.hash_file(src),
		}

	def render(self, property_dict):
		src = property_dict["src"]
		aspect = property_dict.get("aspect", 16 / 9)
		max_dimension = property_dict["max_dimension"]

		if aspect >= 1:
			width = max_dimension
			height = max_dimension / aspect
		else:
			width = max_dimension * aspect
			height = max_dimension

		img_data = self._render_plot_png(src, width, height)
		image = {
			"extension":	"png",
			"img_data":		img_data,
		}
		return image

if __name__ == "__main__":
	from pyradium.RendererCache import RendererCache
	renderer = RendererCache(PlotRenderer())
	print(renderer.render({
		"src":	"/tmp/x.gnuplot",
		"max_dimension": 1920,
	}))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import hashlib
import subprocess
import logging
import xml.dom.minidom
import pyradium
from .Tools import XMLTools
from .TOC import TOCElement, TOCDirective
from .Slide import RenderSlideDirective
from .Acronyms import AcronymDirective
from .Exceptions import XMLFileNotFoundException, MalformedXMLInputException, MalformedJSONInputException

_log = logging.getLogger(__spec__.name)

class Presentation():
	_NAMESPACES = {
		"https://github.com/johndoe31415/pyradium":		"s",
	}

	def __init__(self, meta, content, sources):
		self._meta = meta
		self._content = content
		self._sources = sources
		self._validate_metadata()

	@property
	def meta(self):
		return self._meta

	@property
	def content(self):
		return self._content

	@property
	def sources(self):
		return self._sources

	@classmethod
	def _merge_metadata(cls, meta_dict, injected_dict):
		if not isinstance(injected_dict, dict):
			raise MalformedJSONInputException(f"Could not inject metadata (not a dictionary): {str(injected_dict)}")
		meta_dict.update(injected_dict)
		return meta_dict

	@classmethod
	def parse_xml(cls, filename):
		try:
			dom = xml.dom.minidom.parse(filename)
		except FileNotFoundError as e:
			raise XMLFileNotFoundException(f"Cannot parse {filename}: {str(e)}") from e
		except xml.parsers.expat.ExpatError as e:
			raise MalformedXMLInputException(f"Cannot parse {filename}: {str(e)}") from e
		XMLTools.normalize_ns(dom.documentElement, cls._NAMESPACES)
		presentation = XMLTools.child_tagname(dom, "presentation")
		if presentation is None:
			raise MalformedXMLInputException("No 'presentation' node is present as top node of the XML input document.")
		return (dom, presentation)

	@classmethod
	def load_from_file(cls, filename, rendering_parameters = None):
		(dom, presentation) = cls.parse_xml(filename)
		meta = None
		content = [ ]
		sources = [ filename ]
		for child in presentation.childNodes:
			if child.nodeType != child.ELEMENT_NODE:
				continue

			if child.tagName == "meta":
				meta = XMLTools.xml_to_dict(XMLTools.child_tagname(dom, ("presentation", "meta")))
			elif child.tagName == "slide":
				if not XMLTools.get_bool_attr(child, "hide"):
					content.append(RenderSlideDirective(child))
			elif child.tagName == "include":
				src = child.getAttribute("src")
				if rendering_parameters is not None:
					sub_presentation_filename = rendering_parameters.include_dirs.lookup(src)
					sub_presentation = cls.load_from_file(sub_presentation_filename)
					content += sub_presentation.content
					sources += sub_presentation.sources
				else:
					_log.warning("Ingored include directive because no include directories are known: %s", src)
			elif child.tagName in [ "chapter", "section", "subsection" ]:
				toc_element = TOCElement(child.tagName)
				toc_directive = TOCDirective(toc_element, XMLTools.inner_text(child))
				content.append(toc_directive)
			elif child.tagName == "acronyms":
				content.append(AcronymDirective(child))
			else:
				_log.warning("Ignored unknown tag '%s'.", child.tagName)
		if (rendering_parameters is not None) and (rendering_parameters.injected_metadata is not None):
			meta = cls._merge_metadata(meta, rendering_parameters.injected_metadata)
		return cls(meta, content, sources)

	def _validate_metadata(self):
		pass

	def _determine_sha256(self, filename):
		with open(filename, "rb") as f:
			return hashlib.sha256(f.read()).hexdigest()

	def _determine_git_commit(self, filename):
		try:
			return subprocess.check_output([ "git", "rev-list", "-1", "--all", filename ], stderr = subprocess.DEVNULL).decode().rstrip("\n")
		except subprocess.CalledProcessError:
			return None

	def _determine_version(self, filename):
		return {
			"sha256":	self._determine_sha256(filename),
			"git":		self._determine_git_commit(filename),
		}

	@property
	def version_information(self):
		return [
			{
				"filename":	filename,
				"version":	self._determine_version(filename),
			} for filename in self.sources ]

	@property
	def meta_info(self):
		return {
			"renderer":			pyradium.VERSION,
			"source_versions":	self.version_information,
		}

	def __iter__(self):
		return iter(self._content)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import textwrap
from .Exceptions import InvalidStyleParameterDefinitionException, InvalidStyleParameterValueException

class ParsedStyleParameters():
	def __init__(self, parameters):
		self._parameters = parameters
		self._values = { }

	@property
	def values(self):
		return self._values

	def set_value(self, key, value):
		self._values[key] = value

	def parse(self, text):
		return self._parameters.parse(self, text)

class StyleParameters():
	def __init__(self, definition):
		assert(isinstance(definition, dict))
		self._defs = definition

	def new_parser(self):
		parser = ParsedStyleParameters(self)
		for (name, definition) in self._defs.items():
			if "default" in definition:
				parser.set_value(name, definition["default"])
		return parser

	def parse(self, parser, text):
		split_text = text.split("=", maxsplit = 1)
		if len(split_text) != 2:
			raise InvalidStyleParameterValueException("Expected a key=value tuple for the style parameter, but got: %s" % (text))
		(key, value) = split_text

		key = key.strip()
		value = value.strip()
		if key not in self._defs:
			raise InvalidStyleParameterValueException("Not a valid style parameter: %s (must be one of %s)" % (key, ", ".join(sorted(self._defs))))

		parameter_def = self._defs[key]
		if "type" in parameter_def:
			parameter_type = parameter_def["type"]
			if parameter_type == "str":
				pass
			elif parameter_type == "int":
				try:
					value = int(value)
				except ValueError as e:
					raise InvalidStyleParameterValueException("Not a valid integer value for parameter %s: %s (%s)" % (key, value, str(e))) from e
			elif parameter_type == "float":
				try:
					value = float(value)
				except ValueError as e:
					raise InvalidStyleParameterValueException("Not a valid floating point value for parameter %s: %s (%s)" % (key, value, str(e))) from e
			elif parameter_type == "choice":
				if "choices" not in parameter_def:
					raise InvalidStyleParameterDefinitionException("Choice defined for parameter %s, but 'choices' entry missing." % (key))
				choices = parameter_def["choices"]
				if not isinstance(choices, list):
					raise InvalidStyleParameterDefinitionException("Choice defined for parameter %s, but choices are not a list." % (key))
				if value not in choices:
					raise InvalidStyleParameterValueException("Not a valid value for parameter %s: %s (should be one of %s)" % (key, value, ", ".join(choices)))
			else:
				raise InvalidStyleParameterDefinitionException("Unknown type defined for parameter %s: %s" % (key, parameter_type))
		parser.set_value(key, value)

	def print(self):
		print("Known style parameters:")
		for (name, definition) in sorted(self._defs.items()):
			if definition["type"] == "choice":
				value = "{%s}" % (",".join(definition["choices"]))
			else:
				value = "[%s]" % (definition.get("type", "str"))
			command_name = "%s %s" % (name, value)

			for description_line in textwrap.wrap(definition.get("description", "no description available"), width = 56):
				print("    %-15s    %s" % (command_name, description_line))
				command_name = ""

if __name__ == "__main__":
	sp = StyleParameters({
		"font-style": {
			"description": "influences the font choice used as the main slide text",
			"type":	"choice",
			"choices": [ "normal", "light" ],
			"default": "normal",
		},
		"intval": {
			"type":	"int",
		},
		"scale": {
			"type":	"float",
		}
	})

	parser = sp.new_parser()
	parser.parse("font-style=normal")
	parser.parse("scale=1.234")
	parser.parse("intval=1234")
	print(parser.values)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import InnerTextHook, XMLHookRegistry

@XMLHookRegistry.register_hook
class TerminalHook(InnerTextHook):
	_TAG_NAME = "term"

	@classmethod
	def handle_text(cls, text, rendered_presentation, node):
		replacement_node = node.ownerDocument.createElement("pre")
		replacement_node.setAttribute("class", "terminal")
		if node.hasAttribute("prompt"):
			prompt = node.getAttribute("prompt")
			lines = text.splitlines(keepends = True)
			while len(lines) > 0:
				line = lines.pop(0)
				if not line.startswith(prompt):
					replacement_node.appendChild(node.ownerDocument.createTextNode(line))
					continue
				command = line[len(prompt) : ]
				replacement_node.appendChild(node.ownerDocument.createTextNode(prompt))
				while line.strip().endswith("\\"):
					line = lines.pop(0)
					command += line

				# If there is a trailing newline, we do not want this inside
				# the <span> or otherwise the pasted output on the terminal
				# will contain it too
				has_trailing_nl = command.endswith("\n")
				command = command.rstrip("\n")
				command_node = node.ownerDocument.createElement("span")
				command_node.setAttribute("class", "command")
				command_node.appendChild(node.ownerDocument.createTextNode(command))
				replacement_node.appendChild(command_node)

				if has_trailing_nl:
					# In this case we need to insert the newline back into the
					# "normal" (non-copied) text.
					replacement_node.appendChild(node.ownerDocument.createTextNode("\n"))
		else:
			replacement_node.appendChild(node.ownerDocument.createTextNode(text))

		if node.hasAttribute("height"):
			replacement_node.setAttribute("style", "height: %s" % (node.getAttribute("height")))
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import sys
import xml.dom.minidom

class Helper():
	def __init__(self):
		pass

	@classmethod
	def print_tag(cls, tagname, text):
		doc = xml.dom.minidom.Document()
		node = doc.createElement("raw")
		node.setAttributeNS("xmls", "xmlns:s", "https://github.com/johndoe31415/pyradium")
		child = doc.createElement(tagname)
		child.appendChild(doc.createTextNode(text))
		node.appendChild(child)
		doc.appendChild(node)
		doc.writexml(sys.stdout)

	@classmethod
	def print_tt(cls, text):
		return cls.print_tag("s:tt", text)

	@classmethod
	def print_tex(cls, text):
		return cls.print_tag("s:tex", text)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import tempfile
import subprocess
import logging
import collections
from pyradium.CmdlineEscape import CmdlineEscape
from pyradium.Exceptions import InvalidTeXException
from .BaseRenderer import BaseRenderer

_log = logging.getLogger(__spec__.name)

_TEX_TEMPLATE = r"""
\documentclass[preview,border=1mm,varwidth=true]{standalone}
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage{lmodern}
\usepackage{amsmath}
\usepackage{amssymb}
\begin{document}
%(content)s
\end{document}
"""

@BaseRenderer.register
class LatexFormulaRenderer(BaseRenderer):
	_NAME = "latex"
	_Baseline = collections.namedtuple("Baseline", [ "image_width", "image_height", "upper", "lower", "mid" ])

	def __init__(self, rendering_dpi = 600):
		super().__init__()
		self._rendering_dpi = rendering_dpi
		self._debug_draw_baselines = False

	@property
	def properties(self):
		return {
			"version":			3,
			"rendering_dpi":	self._rendering_dpi,
		}

	def _get_baseline_info(self, png_filename, xoffset):
		# Need 2 pixel wide sample so that ImageMagick does not return an empty image
		crop_info = json.loads(subprocess.check_output([ "convert", "%s[2x+%d+0]" % (png_filename, xoffset), "-trim", "json:-" ]))

		image_height = crop_info[0]["image"]["pageGeometry"]["height"]
		image_width = crop_info[0]["image"]["pageGeometry"]["width"]
		upper_baseline_y_from_top = crop_info[0]["image"]["pageGeometry"]["y"]
		lower_baseline_y_from_top = upper_baseline_y_from_top + crop_info[0]["image"]["geometry"]["height"]

		# Do not choose the exact average, but skewed closer to the lower
		# baseline
		baseline_y_from_top = round((upper_baseline_y_from_top + 7 * lower_baseline_y_from_top) / 8)
		return self._Baseline(image_width = image_width, image_height = image_height, upper = upper_baseline_y_from_top, lower = lower_baseline_y_from_top, mid = baseline_y_from_top)

	def render(self, property_dict):
		with tempfile.TemporaryDirectory(prefix = "pyradium_formula_") as tex_dir:
			# Formula to PDF first using pdflatex
			tex_filename = tex_dir + "/formula.tex"
			pdf_filename = tex_dir + "/formula.pdf"
			png_filename = tex_dir + "/formula.png"

			# Crop 3mm off the left side (1mm baseline bar + 2mm space)
			baseline = r"\rule{1mm}{1pt} \hspace{2mm}"
			left_crop_pixel = round((3 / 25.4) * self._rendering_dpi)
			left_crop_pixel_safe = round((2 / 25.4) * self._rendering_dpi)
			eval_baseline_at_x = round((0.5 / 25.4) * self._rendering_dpi)

			if property_dict.get("long", False):
				content = r"\[" + baseline + property_dict["formula"] + r" \]"
			else:
				content = r"$" + baseline + property_dict["formula"] + r"$"
			with open(tex_filename, "w") as tex_file:
				tex_file.write(_TEX_TEMPLATE % { "content": content })
			_log.debug("Rendering LaTeX formula: %s in directory %s", content, tex_dir)
			try:
				subprocess.check_call([ "pdflatex", "-interaction=nonstopmode", "-output-directory=%s" % (tex_dir), tex_filename ], stdout = _log.subproc_target, stderr = _log.subproc_target)
			except subprocess.CalledProcessError as e:
				raise InvalidTeXException("Invalid TeX in source: %s" % property_dict["formula"]) from e

			# Then render the PDF to PNG
			cmd = [ "convert", "-define", "profile:skip=ICC", "-density", str(self._rendering_dpi), "-trim", "+repage", pdf_filename, png_filename ]
			_log.trace("Converting PDF to PNG in %d dpi: %s", self._rendering_dpi, CmdlineEscape().cmdline(cmd))
			subprocess.check_call(cmd, stdout = _log.subproc_target, stderr = _log.subproc_target)

			_log.trace("Crop on left side: %d (choosing %d to be on safe side); evaluating baseline at x = %d; filename %s", left_crop_pixel, left_crop_pixel_safe, eval_baseline_at_x, png_filename)
			baseline = self._get_baseline_info(png_filename, eval_baseline_at_x)
			_log.trace("Baseline Y from top %d px upper, %d px lower, %d px mid (equals %d px mid from bottom)", baseline.upper, baseline.lower, baseline.mid, baseline.image_height - baseline.mid)

			# Then crop the image finally and capture cropping metadata along the way
			crop_meta = json.loads(subprocess.check_output([ "convert", "-crop", "+%d+0" % (left_crop_pixel_safe), "-trim", png_filename, "json:-" ]))[0]
			cmd = [ "convert", "-crop", "+%d+0" % (left_crop_pixel_safe), "-trim", "+repage", "-strip", png_filename, "png:-" ]
			png_data = subprocess.check_output(cmd)
			_log.trace("Final crop to output size: %s", CmdlineEscape().cmdline(cmd))
			if _log.isEnabledFor(logging.SINGLESTEP):
				with open(tex_dir + "/processed_01_crop_meta.json", "w") as f:
					json.dump(crop_meta, f, indent = 4, sort_keys = True)
				with open(tex_dir + "/processed_01_cropped.png", "wb") as f:
					f.write(png_data)

			_log.trace("Crop upper left corner is at %d, %d and cropped size is %d x %d px", crop_meta["image"]["pageGeometry"]["x"], crop_meta["image"]["pageGeometry"]["y"], crop_meta["image"]["geometry"]["width"], crop_meta["image"]["geometry"]["height"])

			# For debugging purposes, draw the baseline on the image
			if self._debug_draw_baselines:
				cmd = [ "convert", "-stroke", "red", "-draw", "line 0,%d %d,%d" % (baseline.mid, crop_meta["image"]["geometry"]["width"], baseline.mid) ]
				cmd += [ "-", "png:-" ]
				png_data = subprocess.check_output(cmd, input = png_data)
				with open(tex_dir + "/processed_02_baseline.png", "wb") as f:
					f.write(png_data)

			# Compute the shifted baseline in the cropped image
			baseline_from_top_cropped = baseline.mid - crop_meta["image"]["pageGeometry"]["y"]
			baseline_from_bottom_cropped = crop_meta["image"]["geometry"]["height"] - baseline_from_top_cropped
			_log.trace("Adapted baseline offsets for cropped image: %d px from top (equals %d px from bottom).", baseline_from_top_cropped, baseline_from_bottom_cropped)

			# Return an image object
			image = {
				"png_data":	png_data,
				"info": {
					"width": crop_meta["image"]["geometry"]["width"],
					"height": crop_meta["image"]["geometry"]["height"],
					"baseline": baseline_from_bottom_cropped,
				},
			}
			if _log.isEnabledFor(logging.SINGLESTEP):
				_log.singlestep("Interrupting execution.")
				input("Press RETURN to continue...")
			return image



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import logging
import mako.lookup
import mako.exceptions
import markupsafe
import pyradium
from pyradium.Controller import ControllerManager
from pyradium.renderer import BaseRenderer
from .Acronyms import Acronyms
from .RenderedPresentation import RenderedPresentation
from .Exceptions import TemplateErrorException, MalformedStyleConfigurationException, UnknownSlideTypeException
from .Slide import RenderSlideDirective
from .Enums import PresentationFeature
from .Tools import JSONTools
from .StyleParameters import StyleParameters

_log = logging.getLogger(__spec__.name)

class CustomRenderers():
	def __init__(self):
		self._static_instances = {
			"acronym":	Acronyms(),
		}

	def __getitem__(self, name):
		if name in self._static_instances:
			return self._static_instances[name]
		return BaseRenderer.instanciate(name)

class Renderer():
	def __init__(self, presentation, rendering_params):
		self._rendered_slides = [ ]
		self._presentation = presentation
		self._rendering_params = rendering_params
		self._custom_renderers = CustomRenderers()
		self._lookup = mako.lookup.TemplateLookup(list(self._get_mako_lookup_directories()), strict_undefined = True, input_encoding = "utf-8", default_filters = [ "h" ])
		self._template_config_filename = self.lookup_styled_template_file("configuration.json")
		with open(self._template_config_filename) as f:
			self._template_config = json.load(f)
		self._plausibilize_template_config()
		self._ctrlr_mgr = ControllerManager(self)
		self._style_parameters = self._parse_style_parameters()

	def _plausibilize_template_config(self):
		for feature_name in self._template_config.get("dependencies", { }).get("feature", { }):
			try:
				PresentationFeature(feature_name)
			except ValueError as e:
				raise MalformedStyleConfigurationException(f"The stylesheet file {self._template_config_filename} contains an unknown feature in config[\"dependencies\"][\"feature\"]: {feature_name}") from e

	def _parse_style_parameters(self):
		params = StyleParameters(self._template_config.get("parameters", { }))
		parser = params.new_parser()
		for param in self._rendering_params.template_style_opts:
			parser.parse(param)
		return parser.values

	@property
	def presentation(self):
		return self._presentation

	def _get_mako_lookup_directories(self):
		for dirname in self._rendering_params.template_dirs:
			yield dirname
			yield dirname + "/" + self._rendering_params.template_style

	@property
	def rendering_params(self):
		return self._rendering_params

	@property
	def template_config(self):
		return self._template_config

	@property
	def controllers(self):
		return self._ctrlr_mgr

	def get_custom_renderer(self, name):
		return self._custom_renderers[name]

	def lookup_template_file(self, filename):
		return self._rendering_params.template_dirs.lookup(filename)

	def lookup_styled_template_file(self, filename):
		return self._rendering_params.template_dirs.lookup(self._rendering_params.template_style + "/" + filename)

	def lookup_include(self, filename):
		return self._rendering_params.include_dirs.lookup(filename)

	def _determine_slide_types(self):
		slide_types = set()
		for directive in self._presentation:
			if isinstance(directive, RenderSlideDirective):
				slide_types.add(directive.slide_type)
		return slide_types

	def _compute_renderable_slides(self, rendered_presentation):
		renderable_slides = [ ]
		for directive in self._presentation:
			generator = directive.render(rendered_presentation)
			if generator is not None:
				renderable_slides += generator
		return renderable_slides

	def render_file(self, template_filename, rendered_presentation = None, additional_template_args = None):
		def _template_error(text):
			raise TemplateErrorException(text)

		def _jsonify(obj):
			return markupsafe.Markup(json.dumps(JSONTools.round_dict_floats(obj), sort_keys = True, separators = (",", ":")))

		template_args = {
			"pyradium_version":			pyradium.VERSION,
			"renderer":					self,
			"presentation":				self._presentation,
			"template_error":			_template_error,
			"jsonify":					_jsonify,
			"preuri":					self.rendering_params.resource_uri,
			"preuri_ds":				self.rendering_params.resource_uri if self.rendering_params.resource_uri.startswith("/") else ("./" + self.rendering_params.resource_uri),
			"styleopt":					self._style_parameters,
		}
		if rendered_presentation is not None:
			template_args["rendered_presentation"] = rendered_presentation
		if additional_template_args is not None:
			template_args.update(additional_template_args)

		try:
			template = self._lookup.get_template(template_filename)
		except mako.exceptions.MakoException as e:
			raise UnknownSlideTypeException("Could not retrieve template necessary to render slide type %s (searched in: %s)." % (template_filename, ":".join(self._get_mako_lookup_directories()))) from e
		result = template.render(**template_args)
		return result

	def render(self, deploy_directory, resource_directory = None):
		if resource_directory is None:
			resource_directory = deploy_directory
		rendered_presentation = RenderedPresentation(self, deploy_directory = deploy_directory, resource_directory = resource_directory)

		for feature in self.rendering_params.presentation_features:
			rendered_presentation.add_feature(feature)
		_log.trace("Initial feature set: %s", ", ".join(sorted(feature.name for feature in rendered_presentation.features)))

		# Run it first to build the initial TOC and determine feature set
		self._compute_renderable_slides(rendered_presentation)

		_log.debug("Finalized feature set: %s", ", ".join(sorted(feature.name for feature in rendered_presentation.features)))

		# Then copy general dependencies by the used template
		rendered_presentation.handle_dependencies(self._template_config.get("files"))

		# Additionally look for dependencies which are effective because
		# specific slide types are used (e.g., a "quote" slide might need its
		# own CSS)
		for slide_type in self._determine_slide_types():
			slide_type_dependencies = self._template_config.get("dependencies", { }).get("slidetype", { }).get(slide_type)
			if slide_type_dependencies is not None:
				for feature_dependency in slide_type_dependencies.get("features", { }):
					feature_dependency = PresentationFeature(feature_dependency)
					rendered_presentation.add_feature(feature_dependency)
				rendered_presentation.handle_dependencies(slide_type_dependencies)

		# Lastly, add dependencies which are effective because a feature flag
		# is active (e.g., when "interactive" is used, a menu needs to be
		# generated which requires its own CSS)
		for feature in rendered_presentation.features:
			feature_dependency = self._template_config.get("dependencies", { }).get("feature", { }).get(feature.value)
			rendered_presentation.handle_dependencies(feature_dependency)

		# Then run it again to get the page numbers straight (e.g., the TOC
		# pages will be emitted, giving different page numbers). Initialize
		# schedule for second run as well.
		rendered_presentation.finalize_toc()
		rendered_presentation.init_schedule()
		self._compute_renderable_slides(rendered_presentation)
		rendered_presentation.finalize_toc()

		# Compute the schedule before the last run
		rendered_presentation.schedule.compute()

		# Third and final slide run
		for renderable_slide in self._compute_renderable_slides(rendered_presentation):
			additional_template_args = {
				"slide":			renderable_slide,
			}
			template_filename = "slide_%s.html" % (renderable_slide.slide_type)
			rendered_slide = self.render_file(template_filename, rendered_presentation = rendered_presentation, additional_template_args = additional_template_args)
			rendered_presentation.append_slide(rendered_slide)

		rendered_index = self.render_file("base/index.html", rendered_presentation = rendered_presentation)
		rendered_presentation.add_file(self.rendering_params.index_filename, rendered_index, to_deployment_dir = True)

		return rendered_presentation



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools

@XMLHookRegistry.register_hook
class DigitalTimingDiagramHook(BaseHook):
	_TAG_NAME = "dtg"

	@classmethod
	def handle(cls, rendered_presentation, node):
#		if not node.hasAttribute("cmd"):
#			raise MalformedXMLInputException("An s:exec node must have a 'cmd' attribute.")

		properties = {
			"data":	XMLTools.inner_text(node),
		}
		dtg_renderer = rendered_presentation.renderer.get_custom_renderer("dtg")
		result = dtg_renderer.render(properties)

		replacement_node = node.ownerDocument.createElement("s:img")
		replacement_node.setAttribute("value", result.data["svg"])
		replacement_node.setAttribute("filetype", "svg")
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2021-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.Controller import BaseController

class AcronymController(BaseController):
	@property
	def acronyms_per_slide(self):
		return self._options["acronyms_per_slide"]

	@property
	def chars_per_line(self):
		return self._options.get("chars_per_line", 45)

	def render(self):
		acronyms = self.rendered_presentation.renderer.get_custom_renderer("acronym")
		used_acronyms = list(acronyms.get_all_used_acronyms())
		pages = (len(used_acronyms) + self.acronyms_per_slide - 1) // self.acronyms_per_slide
		additional_slide_var_list = [ ]

		current_page = [ ]
		used_lines = 0
		for acronym in used_acronyms:
			acronym_len = len(acronym.acronym) + 2 + len(acronym.text)
			lines = (acronym_len + self.chars_per_line - 1) // self.chars_per_line
			if (used_lines > 0) and (used_lines + lines > self.acronyms_per_slide):
				# Emit a new page
				used_lines = 0
				additional_slide_var_list.append({ "acronyms": current_page })
				current_page = [ ]
			used_lines += lines
			current_page.append(acronym)

		if len(current_page) > 0:
			# Emit the last page
			additional_slide_var_list.append({ "acronyms": current_page })

		yield from self.slide.emit_nocontent_slide(self.rendered_presentation, self.content_containers, additional_slide_var_list)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import string
import contextlib
import collections
from pyradium.Exceptions import InvalidBooleanExpressionException

ParsedElement = collections.namedtuple("ParsedElement", [ "etype", "content" ])

class SpecialCharacter(enum.IntEnum):
	LogicalOr = 0
	LogicalAnd = 1

class BooleanExpressionParser():
	_SPECIAL_CHARS = {
		"|":	SpecialCharacter.LogicalOr,
		"+":	SpecialCharacter.LogicalOr,
		"&":	SpecialCharacter.LogicalAnd,
		"*":	SpecialCharacter.LogicalAnd,
	}
	_LITERAL_CHARS = set(string.ascii_uppercase + string.ascii_lowercase)
	_LONG_LITERAL_CHARS = set(string.ascii_uppercase + string.ascii_lowercase + string.digits + "_")
	_TEXT_CHARS = set(" =," + string.digits)

	def __init__(self, string):
		self._index = 0
		self._string = string
		self._result = None
		self._current = None

	@property
	def char(self):
		if self._index < len(self._string):
			return self._string[self._index]
		else:
			return None

	@property
	def remaining(self):
		return self._string[self._index : ]

	def _produce(self, etype, content):
		element = ParsedElement(etype = etype, content = content)
		if self._result is not None:
			self._current.append(element)
		else:
			self._result = element
			self._current = self._result.content

	def _advance(self):
		self._index += 1

	@contextlib.contextmanager
	def _nested(self, etype):
		if etype is not None:
			content = [ ]
			self._produce(etype, content)
			(prev, self._current) = (self._current, content)
			yield
			self._current = prev
		else:
			yield

	def _parse_inverted(self):
		if self.char == "!":
			self._advance()
			with self._nested("invert"):
				if not self._parse_short_literal() and not self._parse_escaped_long_literal() and not self._parse_parenthesis():
					raise InvalidBooleanExpressionException("Invalid token after '!' sign: {self.remaining}")
			return True
		else:
			return False

	def _parse_long_literal(self):
		result = ""
		while self.char in self._LONG_LITERAL_CHARS:
			result += self.char
			self._advance()
		if result == "":
			return False
		else:
			self._produce("literal", result)
			return True

	def _parse_enclosed(self, etype, start, end, inner_parse_function):
		if self.char == start:
			self._advance()
			with self._nested(etype):
				inner_parse_function()
				if self.char != end:
					raise InvalidBooleanExpressionException(f"Enclosed expression started with '{start}' but does not terminate with '{end}' but '{self.char}' instead.")
				self._advance()
				return True
		else:
			return False

	def _parse_escaped_long_literal(self):
		return self._parse_enclosed(etype = None, start = "[", end = "]", inner_parse_function = self._parse_long_literal)

	def _parse_parenthesis(self):
		return self._parse_enclosed(etype = "parenthesis", start = "(", end = ")", inner_parse_function = self._parse_expression)

	def _parse_simple_char(self, etype, permissible_set, lookup = None):
		if self.char in permissible_set:
			if lookup is None:
				self._produce(etype, self.char)
			else:
				self._produce(etype, lookup(self.char))
			self._advance()
			return True
		return False

	def _parse_short_literal(self):
		return self._parse_simple_char("literal", self._LITERAL_CHARS)

	def _parse_special_char(self):
		return self._parse_simple_char("special", self._SPECIAL_CHARS, lookup = lambda c: self._SPECIAL_CHARS[c])

	def _parse_text(self):
		return self._parse_simple_char("text", self._TEXT_CHARS)

	def _parse_expression(self):
		with self._nested("expr"):
			have_terms = False
			while (self._index < len(self._string)) and (self._parse_escaped_long_literal() or self._parse_short_literal() or self._parse_inverted() or self._parse_special_char() or self._parse_text() or self._parse_parenthesis()):
				have_terms = True
		if not have_terms:
			raise InvalidBooleanExpressionException("Expression has no terms")
		return have_terms

	def parse(self):
		self._parse_expression()
		try:
			if len(self._string) != self._index:
				raise InvalidBooleanExpressionException(f"Trailing unparsed data: {self.remaining}")
		except InvalidBooleanExpressionException as e:
			raise InvalidBooleanExpressionException(f"Error parsing Boolean formula {self._string}: {str(e)}")
		return self._result


class TexBooleanExpressionPrinter():
	def __init__(self, invert_by_overline = True):
		self._invert_by_overline = invert_by_overline

	def _print(self, node, sibling = None):
		if isinstance(node, list):
			prev = None
			for child in node:
				yield from self._print(child, sibling = prev)
				prev = child
			return
		match node.etype:
			case "literal":
				yield r"\textnormal{"
				if "_" not in node.content:
					yield node.content
				else:
					(prefix, suffix) = node.content.split("_", maxsplit = 1)
					yield prefix
					yield "}_{"
					yield suffix
				yield "}"

			case "text":
				yield node.content

			case "invert":
				if self._invert_by_overline:
					if (sibling is not None) and (sibling.etype == "invert"):
						yield r"\ "
					yield r"\overline{"
					if node.content[0].etype == "parenthesis":
						# Don't display parenthesis
						yield from self._print(node.content[0].content)
					else:
						yield from self._print(node.content)
					yield "}"
				else:
					yield r"\neg "
					yield from self._print(node.content)

			case "special":
				yield {
					SpecialCharacter.LogicalOr:		r"\vee",
					SpecialCharacter.LogicalAnd:	r"\wedge",
				}[node.content]

			case "parenthesis":
				yield "("
				yield from self._print(node.content)
				yield ")"

			case "expr":
				yield from self._print(node.content)

			case _:
				raise NotImplementedError(node.etype)

	def print(self, tree):
		return "".join(self._print(tree))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class StyleDict():
	def __init__(self, style_dict: dict | None = None):
		assert((style_dict is None) or isinstance(style_dict, dict))
		self._style = style_dict if (style_dict is not None) else { }

	@classmethod
	def from_style_str(cls, style_str: str):
		style_dict = { }
		for style_item in style_str.split(";"):
			style_item = style_item.strip()
			if style_item == "":
				continue
			style_item = style_item.split(":", maxsplit = 1)
			if len(style_item) == 2:
				(key, value) = style_item
				key = key.strip()
				value = value.strip()
				style_dict[key] = value
		return cls(style_dict)

	@classmethod
	def from_node(cls, node):
		return cls.from_style_str(node.getAttribute("style"))

	def to_node(self, node):
		if (len(self._style) == 0) and node.hasAttribute("style"):
			node.removeAttribute("style")
		else:
			node.setAttribute("style", self.serialize())

	@property
	def is_visible(self):
		if "display" in self._style:
			hidden = (self["display"] == "none")
			return not hidden
		return True

	def serialize(self):
		return ";".join("%s:%s" % (key, value) for (key, value) in self._style.items())

	def __setitem__(self, key: str, value: str):
		self._style[key] = value

	def __getitem__(self, key: str):
		return self._style.get(key)

	def __str__(self):
		return f"StyleDict<{self._style}>"



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import contextlib
import xml.dom.minidom
from pyradium.StyleDict import StyleDict
from pyradium.Tools import XMLTools

class Vector2D():
	def __init__(self, x = 0, y = 0):
		self._x = x
		self._y = y

	@property
	def x(self):
		return self._x

	@property
	def y(self):
		return self._y

	def __iadd__(self, vector):
		return Vector2D(self.x + vector.x, self.y + vector.y)

class SVGStyle():
	def __init__(self, node):
		self._node = node
		if node.hasAttribute("style"):
			self._style_dict = StyleDict.from_style_str(node.getAttribute("style"))
		else:
			self._style_dict = StyleDict()

	def __getitem__(self, key):
		return self._style_dict[key]

	def __setitem__(self, key, value):
		self._style_dict[key] = value
		self._style_dict.to_node(self._node)

	def update(self, style_dict: dict):
		for (key, value) in style_dict.items():
			self._style_dict[key] = value
		self._style_dict.to_node(self._node)

	@classmethod
	def default_path(cls, node):
		style = cls(node)
		style.update({
			"fill":				"none",
			"stroke":			"#000000",
			"stroke-width":		"1px",
			"stroke-linecap":	"butt",
			"stroke-linejoin":	"miter",
			"stroke-opacity":	"1",
		})
		return style

	@classmethod
	def default_text(cls, node):
		style = cls(node)
		style.update({
			"font-style":		"normal",
			"font-weight":		"normal",
			"font-size":		"12px",
			"line-height":		"1.25",
			"font-family":		"sans-serif",
			"white-space":		"pre",
			"fill":				"#000000",
			"fill-opacity":		"1",
			"stroke":			"none",
		})
		return style

class SVGPath():
	def __init__(self, path_node):
		self._node = path_node
		self._style = SVGStyle.default_path(self._node)
		self._cmds = [ ]
		self._pos = Vector2D()

	@property
	def pos(self):
		return self._pos

	@property
	def style(self):
		return self._style

	@contextlib.contextmanager
	def returnto(self):
		prev = self._pos
		yield
		self.move_to(prev.x, prev.y)

	def _append(self, *cmds):
		self._cmds += cmds
		self._node.setAttribute("d", " ".join(self._cmds))

	def move_to(self, x, y):
		self._append("M", f"{x},{y}")
		self._pos = Vector2D(x, y)
		return self

	def move_rel(self, x, y):
		self._append("m", f"{x},{y}")
		self._pos += Vector2D(x, y)
		return self

	def line_to(self, x, y):
		self._append("L", f"{x},{y}")
		self._pos += Vector2D(x, y)
		return self

	def line_rel(self, x, y):
		self._append("l", f"{x},{y}")
		self._pos += Vector2D(x, y)
		return self

	def horiz_rel(self, x):
		self._append("h", f"{x}")
		self._pos += Vector2D(x, 0)
		return self

	def vert_rel(self, y):
		self._append("v", f"{y}")
		self._pos += Vector2D(0, y)
		return self

class SVGText():
	def __init__(self, text_node):
		self._node = text_node
		self._node.setAttribute("xml:space", "preserve")
		self._textdata_node = None
		try:
			self._tspan_node = XMLTools.findfirst(self._node, "tspan")
			try:
				self._textdata_node = next(XMLTools.findall_text(self._tspan_node))
			except StopIteration:
				pass
		except StopIteration:
			self._tspan_node = self._node.appendChild(self._node.ownerDocument.createElement("tspan"))
		if self._textdata_node is None:
			self._textdata_node = self._tspan_node.appendChild(self._node.ownerDocument.createTextNode(""))
		self._style = SVGStyle.default_text(self._node)
		self._tspan_style = SVGStyle(self._tspan_node)

	@property
	def text(self):
		return self._textdata_node.wholeText

	@text.setter
	def text(self, value):
		self._textdata_node.replaceWholeText(value)

	@property
	def style(self):
		return self._style

	@property
	def tspan_style(self):
		return self._tspan_style

	@property
	def font(self):
		return self.tspan_style["font-family"]

	@font.setter
	def font(self, value):
		self.tspan_style["font-family"] = value

class SVGWriter():
	_NAMESPACES = {
		"inkscape": "http://www.inkscape.org/namespaces/inkscape",
		"sodipodi": "http://sodipodi.sourceforge.net/DTD/sodipodi-0.dtd",
		"svg": "http://www.w3.org/2000/svg",
		"rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
		"cc": "http://creativecommons.org/ns#",
		"dc": "http://purl.org/dc/elements/1.1/",
	}

	def __init__(self):
		self._doc = xml.dom.minidom.Document()
		self._root = self._doc.createElement("svg")
		self._root.setAttribute("xmlns", self._NAMESPACES["svg"])
		for (nsname, nsvalue) in self._NAMESPACES.items():
			self._root.setAttribute(f"xmlns:{nsname}", nsvalue)
		self._doc.appendChild(self._root)
		self._defs = None
		self._groups = { }
		self._uid = 0

	def _genid(self):
		self._uid += 1
		return f"id{self._uid}"

	def _add_definition(self, def_node):
		if self._defs is None:
			self._defs = self._doc.createElement("defs")
			self._root.appendChild(self._defs)
		def_id = self._genid()
		def_node.setAttribute("id", def_id)
		self._defs.appendChild(def_node)
		return def_id

	def group(self, group_name):
		if group_name not in self._groups:
			group = self._root.appendChild(self._doc.createElement("g"))
			group.setAttribute("inkscape:groupmode", "layer")
			group.setAttribute("inkscape:label", group_name)
			self._groups[group_name] = group
		return self._groups[group_name]

	def new_path(self, x, y, group_name = "default"):
		path_node = self._doc.createElement("path")
		self.group(group_name).appendChild(path_node)
		svg_path = SVGPath(path_node)
		svg_path.move_to(x, y)
		return svg_path

	def new_text_span(self, x, y, width, height, text, group_name = "default"):
		rect_node = self._doc.createElement("rect")
		rect_node.setAttribute("x", str(x))
		rect_node.setAttribute("y", str(y))
		rect_node.setAttribute("width", str(width))
		rect_node.setAttribute("height", str(height))
		def_id = self._add_definition(rect_node)
		text_node = self.group(group_name).appendChild(self._doc.createElement("text"))
		svg_text = SVGText(text_node)
		svg_text.text = text
		svg_text.style["shape-inside"] = f"url(#{def_id})"
		return svg_text

	def write(self, f):
		self._doc.writexml(f)

	def writefile(self, filename):
		with open(filename, "w") as f:
			self.write(f)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import time
import json
import subprocess
import collections
import logging
import bisect
import urllib.parse
import xml.parsers.expat
import requests
from pyradium.Exceptions import SpellcheckerException

_log = logging.getLogger(__spec__.name)

class LanguageToolConfig():
	def __init__(self, language = "en-US", picky = False, enabled_rules = None, disabled_rules = None, enabled_categories = None, disabled_categories = None):
		self._language = language
		self._picky = picky
		self._enabled_rules = enabled_rules
		self._disabled_rules = disabled_rules
		self._enabled_categories = enabled_categories
		self._disabled_categories = disabled_categories

	@property
	def language(self):
		return self._language

	@property
	def picky(self):
		return self._picky

	@property
	def enabled_rules(self):
		return self._enabled_rules

	@property
	def disabled_rules(self):
		return self._disabled_rules

	@property
	def enabled_categories(self):
		return self._enabled_categories

	@property
	def disabled_categories(self):
		return self._disabled_categories

class SpellcheckerAPI():
	def __init__(self, languagetool_uri, languagetool_config = None):
		self._languagetool_uri = languagetool_uri
		self._languagetool_config = languagetool_config
		if self._languagetool_config is None:
			self._languagetool_config = LanguageToolConfig()
		self._sess = requests.Session()

	def check_data(self, annotations):
		json_payload = {
			"annotation": annotations,
		}
		json_data = json.dumps(json_payload, separators = (",", ":"))
		query_args = {
			"language":	self._languagetool_config.language,
			"data":		json_data,
			"level":	"picky" if self._languagetool_config.picky else "default",
		}
		if self._languagetool_config.enabled_rules is not None:
			query_args["enabledRules"] = ",".join(self._languagetool_config.enabled_rules)
		if self._languagetool_config.disabled_rules is not None:
			query_args["disabledRules"] = ",".join(self._languagetool_config.disabled_rules)
		if self._languagetool_config.enabled_categories is not None:
			query_args["enabledCategories"] = ",".join(self._languagetool_config.enabled_categories)
		if self._languagetool_config.disabled_categories is not None:
			query_args["disabledCategories"] = ",".join(self._languagetool_config.disabled_categories)
		query_str = urllib.parse.urlencode(query_args)
		response = self._sess.post(self._languagetool_uri + "/check?" + query_str, headers = {
			"Accept": "application/json",
		})
		assert(response.status_code == 200)
		return response.json()

	def check_multitext(self, *texts):
		return self.check_data([ { "text": piece } for piece in texts ])

	def check(self, text):
		return self.check_data([{
			"text": text,
		}])

	def try_connect(self, timeout_sec = 3.0, time_between_attempts_sec = 0.25):
		t0 = time.time()
		tend = t0 + timeout_sec
		while time.time() < tend:
			try:
				response = self._sess.get(self._languagetool_uri + "/info")
				if response.status_code == 200:
					return True
			except requests.exceptions.ConnectionError:
				pass
			time.sleep(time_between_attempts_sec)
		return False

class TextChunk():
	def __init__(self, text_type, text, group = None, row = None, column = None):
		assert(text_type in [ "text", "markup" ])
		self._text_type = text_type
		self._text = text
		self._group = group
		self._row = row
		self._column = column

	@property
	def group(self):
		return self._group

	@property
	def row(self):
		return self._row

	@property
	def column(self):
		return self._column

	@property
	def text(self):
		return self._text

	def to_dict(self):
		return { self._text_type: self._text }

	@classmethod
	def joinall(cls, chunks):
		return "".join(chunk.text for chunk in chunks)

	def __str__(self):
		return "Chunk<%s, \"%s\", %s/%s>" % (self._text_type, self._text.replace("\n", r"\n"), self.row, self.column)

class TextChunks():
	def __init__(self):
		self._chunks = [ ]
		self._offsets = [ ]

	@property
	def first_chunk(self):
		return self._chunks[0]

	def append(self, chunk):
		assert(isinstance(chunk, TextChunk))
		if len(self._offsets) == 0:
			start_offset = 0
		else:
			start_offset = self._offsets[-1]
		self._chunks.append(chunk)
		self._offsets.append(start_offset + len(chunk.text))

	def joinall(self):
		return "".join(chunk.text for chunk in self)

	def to_data(self):
		return [ chunk.to_dict() for chunk in self ]

	def dump(self):
		for chunk in self._chunks:
			print(chunk)

	def find_chunk(self, chunk):
		index = self._chunks.index(chunk)
		return self[index]

	def find_offset(self, offset):
		index = bisect.bisect(self._offsets, offset)
		if index >= len(self._offsets):
			index = len(self._offsets) - 1
		return self[index]

	def __getitem__(self, index):
		start_offset = 0 if (index == 0) else self._offsets[index - 1]
		return (self._chunks[index], start_offset)

	def __iadd__(self, others):
		for chunk in others:
			self.append(chunk)
		return self

	def __iter__(self):
		return iter(self._chunks)

class LanguageToolProcess():
	def __init__(self, languagetool_server_jar_filename, port = 12764, languagetool_config = None):
		self._languagetool_server_jar_filename = languagetool_server_jar_filename
		self._port = port
		self._languagetool_config = languagetool_config
		self._proc = None

	def __enter__(self):
		assert(self._proc is None)
		try:
			self._proc = subprocess.Popen([ "java", "-cp", self._languagetool_server_jar_filename, "org.languagetool.server.HTTPServer", "--port", str(self._port), "--allow-origin", "*" ], stdout = _log.subproc_target, stderr = _log.subproc_target)
		except FileNotFoundError as e:
			raise SpellcheckerException("Unable to start LanguageTool on port %d: %s" % (self._port, str(e))) from e

		languagetool_uri = "http://127.0.0.1:%d/v2" % (self._port)
		sc = SpellcheckerAPI(languagetool_uri, self._languagetool_config)
		sc_available = sc.try_connect()
		if not sc_available:
			raise SpellcheckerException("Unable to establish connection to LanguageTool at %s" % (languagetool_uri))
		return sc

	def __exit__(self, *args):
		self._proc.kill()
		self._proc = None

class XMLSpellchecker():
	class ParserMode(enum.IntEnum):
		Normal = 0
		DropText = 1

		# "Markup" does include the text, but if it finds errors in the span,
		# ignores those. Useful for acronyms
		Markup = 2

	_SpellcheckGroup = collections.namedtuple("SpellcheckGroup", [ "description", "identifier", "row", "column", "chunks" ])
	_SpellcheckResult = collections.namedtuple("SpellcheckResult", [ "offense", "chunk", "chunk_offset", "group", "group_offset", "match", "row", "column" ])
	class _GroupIdentifier(enum.IntEnum):
		Slide = 1


	def __init__(self):
		self._parser = xml.parsers.expat.ParserCreate()
		self._parser.StartElementHandler = self._handle_start_element
		self._parser.EndElementHandler = self._handle_end_element
		self._parser.CharacterDataHandler = self._handle_character_data
		self._path = [ ]
		self._mode = [ self.ParserMode.Normal ]
		self._slides = { }
		self._slide_no = 0
		self._metadata = { }
		self._current_group_path = None
		self._current_group = None
		self._groups = [ ]

	@property
	def current_mode(self):
		return self._mode[-1]

	def _add_group(self, group_description, group_identifier):
		(_, column, row) = self._whereami()
		group = self._SpellcheckGroup(description = group_description, identifier = group_identifier, row = row, column = column, chunks = TextChunks())
		self._groups.append(group)
		return group

	def _whereami(self):
		byte_position = self._parser.CurrentByteIndex
		column = self._parser.CurrentColumnNumber + 1
		row = self._parser.CurrentLineNumber
		return (byte_position, column, row)

	def _record_group(self, group_description, text, group_identifier = None):
		(byte_position, column, row) = self._whereami()
		group = self._add_group(group_description, group_identifier)
		group.chunks.append(TextChunk("text", text, group = group, column = column, row = row))

	def _activate_group(self, group_description, group_identifier = None):
		assert(self._current_group is None)
		group = self._add_group(group_description, group_identifier)
		self._current_group = group
		self._current_group_path = list(self._path)

	def _handle_start_element(self, tag, attributes):
		self._path.append(tag)
		next_mode = self.current_mode
		if self._path == [ "presentation", "slide" ]:
			self._slide_no += 1
			self._activate_group("Slide %d content" % (self._slide_no), group_identifier = self._GroupIdentifier.Slide)
		elif (self._path == [ "presentation", "slide", "s:var" ]) and (attributes.get("name") == "heading"):
			self._record_group("Slide %d heading" % (self._slide_no), text = attributes.get("value", ""))
		elif self._path == [ "presentation", "meta", "title" ]:
			self._activate_group("Title metadata")
		elif self._path == [ "presentation", "meta", "subtitle" ]:
			self._activate_group("Subtitle metadata")

		if tag == "br":
			self._handle_character_data("\n")
		elif tag in [ "s:emo", "s:sym", "s:tex", "s:ar", "s:code", "s:term", "s:nsc" ]:
			next_mode = self.ParserMode.DropText
		elif tag == "s:ac":
			next_mode = self.ParserMode.Markup
		elif tag == "s:nth":
			self._handle_character_data("1st")
			next_mode = self.ParserMode.DropText

		self._mode.append(next_mode)

	def _handle_end_element(self, tag):
		if self._current_group_path == self._path:
			self._current_group_path = None
			self._current_group = None
		self._path.pop()
		self._mode.pop()

	def _handle_character_data(self, text):
		if self._current_group is not None:
			(byte_position, column, row) =  self._whereami()
			if self.current_mode == self.ParserMode.Normal:
				self._current_group.chunks.append(TextChunk("text", text, group = self._current_group, row = row, column = column))
			elif self.current_mode == self.ParserMode.Markup:
				self._current_group.chunks.append(TextChunk("markup", text, group = self._current_group, row = row, column = column))

			if self._current_group.identifier == self._GroupIdentifier.Slide:
				if self._slide_no not in self._slides:
					self._slides[self._slide_no] = { "group": self._current_group, "content": [ ] }
				self._slides[self._slide_no]["content"].append(text)

	def parse(self, xml_filename):
		with open(xml_filename, "rb") as f:
			self._parser.ParseFile(f)
		list(self._internal_spellcheck()) # TODO

	def dump(self):
		for group in self._groups:
			print("=> %s" % (group.description))
			for chunk in group.chunks:
				print(chunk)
			print("-" * 120)

	def _internal_spellcheck(self):
		stripped_content = { }
		for (slide_no, content) in self._slides.items():
			stripped = "".join(content["content"])
			stripped = stripped.replace("\n", "").replace("\t", "").replace(" ", "")
			if stripped != "":
				stripped_content[slide_no] = { "group": content["group"], "text": stripped }

		for (slide_no, content) in sorted(stripped_content.items()):
			next_slide_no = slide_no + 1
			if next_slide_no in stripped_content:
				text = content["text"]
				next_text = stripped_content[next_slide_no]["text"]
				if text == next_text:
					match = {
						"message": "Possible duplicate slide %d / %d" % (slide_no, next_slide_no),
						"replacements": [ ],
						"context": {
							"text": "N/A",
						},
						"rule": {
							"id": "pyradium-duplicate-slide-%d-%d" % (slide_no, next_slide_no)
						}
					}
					spellcheck_result = self._SpellcheckResult(offense = None, chunk = None, chunk_offset = None, group = content["group"], group_offset = None, match = match, row = content["group"].row, column = content["group"].column)
					yield spellcheck_result

	def spellcheck(self, spellcheck_api):
		yield from self._internal_spellcheck()

		separator = "\n\n"

		all_chunks = TextChunks()
		for group in self._groups:
			all_chunks += group.chunks
			all_chunks.append(TextChunk("markup", separator))
		result = spellcheck_api.check_data(all_chunks.to_data())

		for match in result["matches"]:
			(chunk, chunk_start_offset) = all_chunks.find_offset(match["offset"])
			chunk_offset = match["offset"] - chunk_start_offset

			group = chunk.group
			(first_chunk, first_chunk_start_offset) = all_chunks.find_chunk(group.chunks.first_chunk)
			group_offset = match["offset"] - first_chunk_start_offset

			row = chunk.row
			column = chunk.column
			for char in chunk.text[:chunk_offset]:
				if char == "\n":
					row += 1
					column = 1
				else:
					column += 1
			offense = chunk.text[chunk_offset : chunk_offset + match["length"]]
			spellcheck_result = self._SpellcheckResult(offense = offense, chunk = chunk, chunk_offset = chunk_offset, group = group, group_offset = group_offset, match = match, row = row, column = column)
			yield spellcheck_result

if __name__ == "__main__":
	ltp = LanguageToolProcess("lt/LanguageTool-5.5/languagetool-server.jar")
	xml_spellcheck = XMLSpellchecker()
	xml_spellcheck.parse("examples/example.xml")
	with ltp as api:
		for result in xml_spellcheck.spellcheck(api):
			offense = result.chunk.text[result.chunk_offset : result.chunk_offset + result.match["length"]]

			group_text = result.group.chunks.joinall()
			offense_group = group_text[result.group_offset : result.group_offset + result.match["length"]]
			assert(offense == offense_group)
			print("%s [line %d, col %d] \"%s\": %s" % (result.group.description, result.row, result.column, offense, result.match["message"]))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2021 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import json
import logging
from .BaseAction import BaseAction
from .RenderingParameters import RenderingParameters
from .Presentation import Presentation

_log = logging.getLogger(__spec__.name)

class ActionDumpMetadata(BaseAction):
	def run(self):
		if self._args.inject_metadata is None:
			injected_metadata = None
		else:
			with open(self._args.inject_metadata) as f:
				injected_metadata = json.load(f)
				_log.debug("Injected metadata: %s", str(injected_metadata))

		rendering_parameters = RenderingParameters(
				include_dirs = [ os.path.dirname(self._args.infile) or "." ] + self._args.include_dir,
				injected_metadata = injected_metadata)
		presentation = Presentation.load_from_file(self._args.infile, rendering_parameters)

		if self._args.pretty_print:
			print(json.dumps(presentation.meta, indent = 4, sort_keys = True))
		else:
			print(json.dumps(presentation.meta, separators = (",", ":")))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.Tools import XMLTools
from pyradium.Exceptions import InvalidTransformationException, MalformedXMLInputException

@XMLHookRegistry.register_hook
class ImgHook(BaseHook):
	_TAG_NAME = "img"

	@classmethod
	def _parse_transformations(cls, node):
		transformations = [ ]

		variables = { }
		for child_node in node.childNodes:
			if (child_node.nodeType == child_node.ELEMENT_NODE) and (child_node.tagName == "s:format"):
				if not (child_node.hasAttribute("name") and child_node.hasAttribute("value")):
					raise InvalidTransformationException("s:format element needs a 'name' and 'value' attribute.")
				name = child_node.getAttribute("name")
				value = child_node.getAttribute("value")

				if child_node.hasAttribute("type"):
					vtype = child_node.getAttribute("type")
					match vtype:
						case "int":
							value = int(value)
						case _:
							raise InvalidTransformationException(f"s:format element recevied unsupported type '{vtype}'.")
				variables[name] = value

		if len(variables) > 0:
			transformations.append({
				"cmd":			"format_text",
				"variables":	variables,
			})
		return transformations

	@classmethod
	def handle(cls, rendered_presentation, node):
		transformations = cls._parse_transformations(node)

		if node.hasAttribute("src") and node.hasAttribute("value"):
			raise MalformedXMLInputException("Image node may not have both 'src' and 'value' attribute.")
		elif not node.hasAttribute("src") and not node.hasAttribute("value"):
			raise MalformedXMLInputException("Image node needs at least a 'src' or 'value' node.")
		elif node.hasAttribute("value") and not node.hasAttribute("filetype"):
			raise MalformedXMLInputException("Image node with a 'value' literal specification also needs a 'filetype' attribute to specify how to interpret the data.")

		properties = {
			"max_dimension":	rendered_presentation.renderer.rendering_params.image_max_dimension,
		}
		if node.hasAttribute("src"):
			properties["src"] = rendered_presentation.renderer.lookup_include(node.getAttribute("src"))
		else:
			# Literal specification as value
			properties["value"] = node.getAttribute("value").encode("utf-8")

		if node.hasAttribute("filetype"):
			properties["filetype"] = node.getAttribute("filetype")

		if len(transformations) > 0:
			properties["svg_transform"] = transformations
		img_renderer = rendered_presentation.renderer.get_custom_renderer("img")
		rendered_image = img_renderer.render(properties)
		local_filename = "imgs/img/%s.%s" % (rendered_image.keyhash, rendered_image.data["extension"])
		uri = "%simgs/img/%s.%s" % (rendered_presentation.renderer.rendering_params.resource_uri, rendered_image.keyhash, rendered_image.data["extension"])

		img_style = [ ]
		if node.hasAttribute("width"):
			img_style.append(("width", node.getAttribute("width")))
		if node.hasAttribute("height"):
			img_style.append(("height", node.getAttribute("height")))
		if node.hasAttribute("render"):
			img_style.append(("image-rendering", node.getAttribute("render")))

		create_filldiv = XMLTools.get_bool_attr(node, "fill", default_value = True)

		img_node = node.ownerDocument.createElement("img")
		img_node.setAttribute("src", uri)
		img_node.setAttribute("class", "fill")
		if len(img_style) > 0:
			img_node.setAttribute("style", ";".join("%s:%s" %  (key, value) for (key, value) in img_style))

		if create_filldiv:
			replacement_node = node.ownerDocument.createElement("div")
			replacement_node.setAttribute("class", "fillimg")
			replacement_node.appendChild(img_node)
		else:
			replacement_node = img_node

		rendered_presentation.add_file(local_filename, rendered_image.data["img_data"])
		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import sys
import json
import base64
import zlib
import logging
import subprocess
from .BaseAction import BaseAction
from .Spellcheck import XMLSpellchecker, SpellcheckerAPI, LanguageToolProcess, LanguageToolConfig
from .SpellcheckDictionary import SpellcheckDictionary
from .CmdlineEscape import CmdlineEscape

_log = logging.getLogger(__spec__.name)

class ActionSpellcheck(BaseAction):
	_EVIM_SEPARATOR = "::"

	def _add_finding_print(self, spellcheck_result):
		msg = [ ]
		msg.append("%s " % (spellcheck_result.group.description))
		msg.append("[line %d, col %d] " % (spellcheck_result.row, spellcheck_result.column))
		if spellcheck_result.offense is not None:
			msg.append("\"%s\": " % (spellcheck_result.offense))
		else:
			msg.append(": ")
		msg.append("%s" % (spellcheck_result.match["message"]))
		if len(spellcheck_result.match["replacements"]) > 0:
			msg.append(" (suggest %s)" % (" or ".join(replacement["value"] for replacement in spellcheck_result.match["replacements"][:3])))
		print("".join(msg), file = self._f)

	def _get_vim_msg(self, spellcheck_result):
		if spellcheck_result.offense is not None:
			msg = [ "\"%s\": %s" % (spellcheck_result.offense, spellcheck_result.match["message"]) ]
		else:
			msg = [ spellcheck_result.match["message"] ]
		if len(spellcheck_result.match["replacements"]) > 0:
			msg.append(" (suggest %s)" % (" or ".join(replacement["value"] for replacement in spellcheck_result.match["replacements"][:3])))
		return "".join(msg)

	def _add_finding_vim(self, spellcheck_result):
		print("%s:%d:%d:%s" % (self._args.infile, spellcheck_result.row, spellcheck_result.column, self._get_vim_msg(spellcheck_result)), file = self._f)

	def _add_finding_evim(self, spellcheck_result):
		data = {
			"offense":	spellcheck_result.offense,
			"ctx":		spellcheck_result.match["context"]["text"],
			"rule":		spellcheck_result.match["rule"]["id"],
		}
		bin_data = zlib.compress(json.dumps(data, sort_keys = True, separators = (",", ":")).encode("ascii"))
		encoded_data = base64.b64encode(bin_data).decode("ascii")
		print(self._EVIM_SEPARATOR.join([ encoded_data, self._args.infile, str(spellcheck_result.row), str(spellcheck_result.column), self._get_vim_msg(spellcheck_result) ]), file = self._f)

	def _add_finding_fulljson(self, spellcheck_result):
		finding = {
			"chunk":			spellcheck_result.chunk.text,
			"chunk_offset":		spellcheck_result.chunk_offset,
			"group": {
				"description":	spellcheck_result.group.description,
				"chunks":		[ chunk.text for chunk in spellcheck_result.group.chunks ]
			},
			"group_offset":		spellcheck_result.group_offset,
			"match":			spellcheck_result.match,
		}
		self._findings.append(finding)

	def _emit_findings_fulljson(self):
		json.dump(self._findings, self._f)

	def _run_spellcheck(self, spellchecker_api):
		self._findings = [ ]
		finding_handler = getattr(self, "_add_finding_" + self._args.mode)
		emission_handler = getattr(self, "_emit_findings_" + self._args.mode, None)
		for spellcheck_result in self._xml_spellchecker.spellcheck(spellchecker_api):
			if not self._dictionary.has_exception(offense = spellcheck_result.offense, context = spellcheck_result.match["context"]["text"], rule_id = spellcheck_result.match["rule"]["id"]):
				finding_handler(spellcheck_result)
		if emission_handler is not None:
			emission_handler()

	def run(self):
		if self._args.vim:
			if not self._args.mode in [ "vim", "evim" ]:
				print("When trying to run in VIM mode, you need to specify either the 'vim' or 'evim' output facility.", file = sys.stderr)
				return 1
			if self._args.outfile is None:
				print("When trying to run in VIM mode, you need to specify an output file, not stdout.", file = sys.stderr)
				return 1

		if (self._args.outfile is not None) and (not self._args.force) and (os.path.exists(self._args.outfile)):
			print("Refusing to overwrite: %s" % (self._args.outfile), file = sys.stderr)
			return 1
		if self._args.outfile is None:
			self._f = sys.stdout
		else:
			self._f = open(self._args.outfile, "w")

		self._dictionary = SpellcheckDictionary.open_global_dict()
		languagetool_config = LanguageToolConfig(language = self._args.language, disabled_rules = [ "WHITESPACE_RULE", "COMMA_PARENTHESIS_WHITESPACE" ])
		try:
			self._xml_spellchecker = XMLSpellchecker()
			self._xml_spellchecker.parse(self._args.infile)

			if self._args.jar is not None:
				ltp = LanguageToolProcess(self._args.jar, languagetool_config = languagetool_config)
				with ltp as spellchecker_api:
					self._run_spellcheck(spellchecker_api)
			else:
				spellchecker_api = SpellcheckerAPI(languagetool_uri = self._args.uri)
				self._run_spellcheck(spellchecker_api)
		finally:
			self._f.close()

		if self._args.vim:
			vim_errorformat = {
				"vim":		r"%f:%l:%c:%m",
				"evim":		self._EVIM_SEPARATOR.join([ r"%[A-Za-z0-9/+=]%\\\\+", r"%f", r"%l", r"%c", r"%m" ]),
			}[self._args.mode]
			cmd = [ "vi", "-c", ":set errorformat=%s" % (vim_errorformat), "-c", ":cf %s" % (self._args.outfile) ]
			_log.info("Running: %s", CmdlineEscape().cmdline(cmd))
			subprocess.check_call(cmd)
		return 0



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from .Acrofile import Acrofile
from .BaseAction import BaseAction

class ActionAcroAdd(BaseAction):
	def run(self):
		acrofile = Acrofile.load_from_file(self._args.acrofile)
		while True:
			acronym = input("Acronym: ")
			if acrofile.has(acronym):
				acro_data = acrofile.get(acronym)
				print("Acronym '%s' already in database." % (acronym))
				for i in range(1, 100):
					suggestion = "%s-%d" % (acronym, i)
					if not acrofile.has(suggestion):
						print("Suggest to use: %s" % (suggestion))
						break
				if "text" in acro_data:
					print("Text   : %s" % (acro_data["text"]))
				if "acronym" in acro_data:
					print("Acronym: %s" % (acro_data["acronym"]))
				if "uri" in acro_data:
					print("URI    : %s" % (acro_data["uri"]))
			else:
				break

		text = input("Text   : ")
		actual_acronym = input("Acronym: ")
		uri = input("URI    : ")
		acro_data = { }
		if text != "":
			acro_data["text"] = text
		if actual_acronym != "":
			acro_data["acronym"] = actual_acronym
		if uri != "":
			acro_data["uri"] = uri

		acrofile.add_entry(acronym, acro_data)
		acrofile.write_to_file(self._args.acrofile)



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import logging
from pyradium.xmlhooks.XMLHookRegistry import BaseHook, XMLHookRegistry
from pyradium.CircuitJS import CircuitJSCircuit
from pyradium.Exceptions import MalformedXMLInputException, FailedToLookupFileException

_log = logging.getLogger(__spec__.name)

@XMLHookRegistry.register_hook
class CircuitHook(BaseHook):
	_TAG_NAME = "circuit"

	@classmethod
	def handle(cls, rendered_presentation, node):
		circuit = CircuitJSCircuit.from_xml(node, find_file_function = rendered_presentation.renderer.lookup_include)

		replacement_node = node.ownerDocument.createElement("a")
		replacement_node.setAttribute("href", circuit.full_url(include_circuit = True))

		name = circuit.get_presentation_parameter("name")
		content = circuit.get_presentation_parameter("content")
		if content is not None:
			match content:
				case "image":
					if name is None:
						raise MalformedXMLInputException("s:circuit requires an image by 'content' parameter, but does not name the circuit.")

					filename = f"circuit_{name}.svg"
					try:
						lookup = rendered_presentation.renderer.lookup_include(filename)
						img = node.ownerDocument.createElement("s:img")
						img.setAttribute("src", filename)
						replacement_node.appendChild(img)
					except FailedToLookupFileException:
						error_text = f"Missing rendered circuit: {name}"
						replacement_node.appendChild(node.ownerDocument.createTextNode(error_text))
						_log.warning(error_text)
				case _:
					raise MalformedXMLInputException(f"s:circuit has unknown 'content' parameter: {content}")
		else:
			# Verbatim take the content nodes
			for child_node in circuit.display_content:
				replacement_node.appendChild(child_node)

		return replacement_node



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import io
import logging
import tempfile
import mimetypes
import subprocess
from pyradium.CmdlineEscape import CmdlineEscape
from pyradium.Tools import ImageTools, HashTools
from pyradium.Exceptions import UsageException, ImageRenderingException
from pyradium.SVGTransformation import SVGTransformation
from .BaseRenderer import BaseRenderer

_log = logging.getLogger(__spec__.name)

@BaseRenderer.register
class ImageRenderer(BaseRenderer):
	_NAME = "img"

	@property
	def properties(self):
		return {
			"version":			1,
		}

	def _render_raw_svg(self, content, max_dimension, svg_transform = None):
		with tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = ".svg", mode = "wb") as input_file, tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = ".png") as output_file:
			input_file.write(content)
			input_file.flush()

			svg_info = ImageTools.get_image_info(input_file.name)
			if svg_info["image"]["geometry"]["width"] > svg_info["image"]["geometry"]["height"]:
				scale_param = "-w"
			else:
				scale_param = "-h"
			cmd = [ "inkscape", "-o", output_file.name, scale_param, str(max_dimension), input_file.name ]
			_log.debug("Rendering SVG: %s", CmdlineEscape().cmdline(cmd))
			subprocess.check_call(cmd, stdout = _log.subproc_target, stderr = _log.subproc_target)
			extension = "png"
			img_data = output_file.read()
		return (extension, img_data)

	def _render_svg(self, content, max_dimension, svg_transform = None):
		if svg_transform is None:
			return self._render_raw_svg(content, max_dimension)
		else:
			# Transform SVG before rendering it
			with tempfile.NamedTemporaryFile(prefix = "pyradium_svg_", suffix = ".svg", mode = "wb") as svg_file:
				# First write the unmodified file to the tempfile
				svg_file.write(content)
				svg_file.flush()

				# Then perform the transformation and write it back
				transformed_svg = io.StringIO()
				svg = SVGTransformation(svg_file.name)
				svg.apply_all(svg_transform)
				svg.write(transformed_svg)

				transformed_content = transformed_svg.getvalue().encode("utf-8")
				return self._render_raw_svg(transformed_content, max_dimension)

	def _render_raster_bitmap(self, content, mimetype, max_dimension):
		# Resize using ImageMagick
		extension = "jpg" if (mimetype == "image/jpeg") else "png"

		with tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = mimetypes.guess_extension(mimetype)) as input_file, tempfile.NamedTemporaryFile(prefix = "pyradium_img_", suffix = "." + extension) as output_file:
			input_file.write(content)
			input_file.flush()

			cmd = [ "convert", "-geometry", ">%dx%d" % (max_dimension, max_dimension), input_file.name, output_file.name ]
			_log.debug("Rendering raster image: %s", CmdlineEscape().cmdline(cmd))
			try:
				subprocess.check_call(cmd, stdout = _log.subproc_target, stderr = _log.subproc_target)
			except subprocess.CalledProcessError as e:
				raise ImageRenderingException(f"Failed to render {input_file.name} while trying to execute: {CmdlineEscape().cmdline(cmd)}") from e
			img_data = output_file.read()
		return (extension, img_data)

	def rendering_key(self, property_dict):
		if "src" in property_dict:
			# Specification as filename
			srchash = HashTools.hash_file(property_dict["src"])
		else:
			# Literal specification as data
			srchash = HashTools.hash_data(property_dict["value"])
		return {
			"srchash":		srchash,
		}

	def _determine_mimetype(self, property_dict):
		if "filetype" in property_dict:
			filename = f"x.{property_dict['filetype']}"
			return mimetypes.guess_type(filename)[0]
		else:
			return mimetypes.guess_type(property_dict["src"])[0]

	def render(self, property_dict):
		max_dimension = property_dict["max_dimension"]
		mimetype = self._determine_mimetype(property_dict)

		if "src" in property_dict:
			with open(property_dict["src"], "rb") as f:
				content = f.read()
		else:
			# Literal content specitication
			content = property_dict["value"]
			assert(isinstance(content, bytes))

		if mimetype == "image/svg+xml":
			(extension, img_data) = self._render_svg(content, max_dimension, svg_transform = property_dict.get("svg_transform"))
		else:
			if "svg_transform" in property_dict:
				raise UsageException(f"SVG transformation requested, but source file does not have SVG mimetype: {mimetype} / {str(property_dict)}")
			(extension, img_data) = self._render_raster_bitmap(content, mimetype, max_dimension)

		image = {
			"extension":	extension,
			"img_data":		img_data,
		}
		return image

if __name__ == "__main__":
	from pyradium.RendererCache import RendererCache
	renderer = RendererCache(ImageRenderer())
	print(renderer.render({
		"max_dimension": 100,
		"src":	"test.svg",
	}))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import unittest
from pyradium.BooleanExpression import BooleanExpressionParser, TexBooleanExpressionPrinter
from pyradium.Exceptions import InvalidBooleanExpressionException

class BooleanExpressionTests(unittest.TestCase):
	def setUp(self):
		self._tex_default = TexBooleanExpressionPrinter()
		self._tex_no_overline = TexBooleanExpressionPrinter(invert_by_overline = False)

	def test_parse_short_literal(self):
		BooleanExpressionParser("A B C").parse()

	def test_parse_long_literal(self):
		BooleanExpressionParser("[ABC][DEF]").parse()

	def test_parse_invert(self):
		BooleanExpressionParser("A!B!C").parse()

	def test_parse_special(self):
		BooleanExpressionParser("A & B | C").parse()

	def test_parse_complete(self):
		BooleanExpressionParser("X = (A!B!C) | !(ABC) | ![ABC] & X!YZ | (A & (B | !C))").parse()

	def test_parse_error_open_parenthesis(self):
		with self.assertRaises(InvalidBooleanExpressionException):
			BooleanExpressionParser("X = (A!B!C").parse()

	def test_parse_error_inverted(self):
		with self.assertRaises(InvalidBooleanExpressionException):
			BooleanExpressionParser("X = !").parse()

	def test_parse_error_empty(self):
		with self.assertRaises(InvalidBooleanExpressionException):
			BooleanExpressionParser("X = ()").parse()

	def test_parse_error_trailiing(self):
		with self.assertRaises(InvalidBooleanExpressionException):
			BooleanExpressionParser("X#").parse()

	def test_tex_default_simple(self):
		tree = BooleanExpressionParser("XYZ").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{X}\textnormal{Y}\textnormal{Z}")

	def test_tex_default_long_literal(self):
		tree = BooleanExpressionParser("[XYZ]").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{XYZ}")

	def test_tex_default_inv_long_literal(self):
		tree = BooleanExpressionParser("![XYZ]").parse()
		self.assertEqual(self._tex_default.print(tree), r"\overline{\textnormal{XYZ}}")

	def test_tex_default_inverted_literal(self):
		tree = BooleanExpressionParser("X!Y!Z").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{X}\overline{\textnormal{Y}}\overline{\textnormal{Z}}")

	def test_tex_default_logical_or(self):
		tree = BooleanExpressionParser("X | Y").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{X} \vee \textnormal{Y}")

	def test_tex_default_logical_and(self):
		tree = BooleanExpressionParser("X & Y").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{X} \wedge \textnormal{Y}")

	def test_tex_default_complex(self):
		tree = BooleanExpressionParser("X = (A!B!C) | !(ABC) | ![ABC] & X!YZ | (A & (B | !C))").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{X} = (\textnormal{A}\overline{\textnormal{B}}\overline{\textnormal{C}}) \vee \overline{\textnormal{A}\textnormal{B}\textnormal{C}} \vee \overline{\textnormal{ABC}} \wedge \textnormal{X}\overline{\textnormal{Y}}\textnormal{Z} \vee (\textnormal{A} \wedge (\textnormal{B} \vee \overline{\textnormal{C}}))")

	def test_tex_default_realworld_1(self):
		tree = BooleanExpressionParser("[CLK] = 1").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{CLK} = 1")

	def test_tex_default_realworld_2(self):
		tree = BooleanExpressionParser("!Q = 1").parse()
		self.assertEqual(self._tex_default.print(tree), r"\overline{\textnormal{Q}} = 1")

	def test_tex_default_realworld_3(self):
		tree = BooleanExpressionParser("Z = !(!A!B!C | ![foo])").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{Z} = \overline{\overline{\textnormal{A}}\overline{\textnormal{B}}\overline{\textnormal{C}} \vee \overline{\textnormal{foo}}}")

	def test_tex_default_realworld_4(self):
		tree = BooleanExpressionParser("[CLK] = AB!C!D | A!B!D & (A | BD) | !(A | BC)").parse()
		self.assertEqual(self._tex_default.print(tree), r"\textnormal{CLK} = \textnormal{A}\textnormal{B}\overline{\textnormal{C}}\overline{\textnormal{D}} \vee \textnormal{A}\overline{\textnormal{B}}\overline{\textnormal{D}} \wedge (\textnormal{A} \vee \textnormal{B}\textnormal{D}) \vee \overline{\textnormal{A} \vee \textnormal{B}\textnormal{C}}")

	def test_tex_no_overline_realworld_4(self):
		tree = BooleanExpressionParser("[CLK] = AB!C!D | A!B!D & (A | BD) | !(A | BC)").parse()
		self.assertEqual(self._tex_no_overline.print(tree), r"\textnormal{CLK} = \textnormal{A}\textnormal{B}\neg \textnormal{C}\neg \textnormal{D} \vee \textnormal{A}\neg \textnormal{B}\neg \textnormal{D} \wedge (\textnormal{A} \vee \textnormal{B}\textnormal{D}) \vee \neg (\textnormal{A} \vee \textnormal{B}\textnormal{C})")



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import subprocess
import logging
import os
import contextlib
import enum
import shutil
import urllib.parse
try:
	import aiohttp.web
except ModuleNotFoundError:
	aiohttp = None
import asyncio
from pyradium.Presentation import Presentation
from pyradium.Tools import XMLTools, FileTools
from pyradium.CircuitJS import CircuitJSCircuit
from pyradium.CmdlineEscape import CmdlineEscape
from pyradium.FileLookup import FileLookup
from .BaseModifyCommand import BaseModifyCommand

_log = logging.getLogger(__spec__.name)
_empty_circuit_urlcomponent = "CQAgjCAMB0l3BWcMBMcUHYMGZIA4UA2ATmIxAUgpABZsKBTAWjDACgg"

class _WriteBack(enum.Enum):
	NoWriteback = "no"
	Inline = "inline"
	ExternalFile = "extfile"

@BaseModifyCommand.register
class CircuitJSRenderImages(BaseModifyCommand):
	_NAME = "circuitjs"
	_DESCRIPTION = "Render CircuitJS circuits"
	_VISIBLE = (aiohttp is not None)

	@classmethod
	def _gen_parser(cls, parser):
		parser.add_argument("-I", "--include-dir", metavar = "path", action = "append", default = [ ], help = "Specifies an additional include directory in which, for example, circuits are located which are referenced from the presentation. Can be issued multiple times.")
		parser.add_argument("-u", "--circuitjs-uri", metavar = "uri", default = "http://127.0.0.1:8123/circuitws.html", help = "URI that the websocket CircuitJS is hosted at. Defaults to %(default)s.")
		parser.add_argument("--websocket-port", metavar = "port", type = int, default = 3424, help = "Run websocket server on localhost at this port. Defaults to %(default)d.")
		parser.add_argument("-c", "--capture-circuit", action = "store_true", help = "Not only capture the SVG files, but also the circuits itself.")
		parser.add_argument("-t", "--settle-time", metavar = "secs", type = float, default = 1.0, help = "Simulate circuit for this duration before capturing it (defaults to %(default).1f seconds).")
		parser.add_argument("--no-xdg-open", action = "store_true", help = "Do not call xdg-open automatically to start the process.")
		parser.add_argument("--no-svg-postprocessing", action = "store_true", help = "Do not call Inkscape to adapt the returned SVG canvas size.")
		parser.add_argument("--no-shutdown", action = "store_true", help = "Keep CircuitJS running after everything is rendered; do not cause it to shut down.")
		parser.add_argument("--only-circuit", metavar = "name", action = "append", default = [ ], help = "Only process circuits with the given name. Can be specified multiple times. If not specified, all circuits are affected.")
		parser.add_argument("-w", "--write-back", metavar = "{" + ",".join(choice.value for choice in _WriteBack) + "}", choices = _WriteBack, type = _WriteBack, default = "no", help = "This option not only reads out the circuit after it has settled, but it also writes those changes back to the XML presentation file. This can either happen inline (directly as text) or as an external file reference.")
		parser.add_argument("-s", "--simulation-change", choices = [ "reset_params" ], action = "append", default = [ ], help = "Manipulate the circuit in a defined way. Right now there is only 'reset_params' supported, which resets timestep/simulation speed to the initial defaults. Can be supplied multiple times.")
		parser.add_argument("-o", "--output-dir", metavar = "path", default = "circuits/", help = "Output directory to store the circuit SVGs into.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increases verbosity. Can be specified multiple times to increase.")
		parser.add_argument("infile", help = "Input XML file of the presentation.")

	async def _on_startup(self, app):
		if not self._args.no_xdg_open:
			self._subproc = subprocess.Popen([ "xdg-open", self._target_uri ], stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL)

	async def _websocket_handler(self, request):
		if self._connection:
			return
		self._connection = True

		_log.debug("Websocket client connected from %s: %s", request.remote, request)
		ws = aiohttp.web.WebSocketResponse()
		await ws.prepare(request)

		async def tx_rx(msg, expect_response = True, wait_for_event = None):
			self._msgid += 1
			used_msgid = self._msgid
			msg["msgid"] = used_msgid
			_log.trace("Sending: %s", msg)
			await ws.send_json(msg)
			while expect_response:
				answer = await ws.receive_json()
				if (wait_for_event is None) and ("msgid" in answer) and (answer["msgid"] == used_msgid):
					_log.trace("Response: %s", answer)
					return answer
				if (wait_for_event is not None) and (answer["type"] == "event") and (answer["code"] == wait_for_event):
					_log.trace("Response event: %s", answer)
					return answer
				else:
					_log.trace("Response with different/no message ID: %s", answer)



		# We do not know if the simulator is available yet. Test it out
		status = await tx_rx({ "cmd": "status" })
		if status["status"] == "error":
			# Initially, simulator is not available yet. Wait for it to become available.
			while True:
				answer = await ws.receive_json()
				if (answer["type"] == "event") and (answer["code"] == "reload_complete"):
					break

		try:
			circuit_params = None
			for (cno, circuit) in enumerate(self._circuits, 1):
				_log.info("Processing circuit %d of %d: %s", cno, len(self._circuits), circuit.get_presentation_parameter("name"))
				if circuit.circuit_params() != circuit_params:
					# TODO: Disregard "ctz" here.
					circuit_params = circuit.circuit_params()
					circuit_params["ctz"] = _empty_circuit_urlcomponent
					_log.debug("Circuit parameters changed, forcing reload of circuit simulator")
					await tx_rx({ "cmd": "reload", "args": circuit_params }, wait_for_event = "reload_complete")

				# Load circuit
				await tx_rx({ "cmd": "circuit_import", "circuit": circuit.circuit_text })

				# Wait for simulation to settle
				# TODO actually determine simulation time
				t0 = await tx_rx({ "cmd": "status" })
				await asyncio.sleep(self._args.settle_time)
				t1 = await tx_rx({ "cmd": "status" })

				# Then make a snapshot of the SVG
				svg = await tx_rx({ "cmd": "get_svg" }, wait_for_event = "svg_rendered")

				# Create the output SVG file
				output_filename_svg = f"{self._args.output_dir}/circuit_{circuit.get_presentation_parameter('name')}.svg"
				with open(output_filename_svg, "w") as f:
					f.write(svg["data"])

				# Postprocess the SVG file
				if not self._args.no_svg_postprocessing:
					cmd = [ "inkscape", "-g", "--verb=FitCanvasToDrawing;FileSave;FileQuit", output_filename_svg ]
					_log.debug("Fitting SVG to canvas: %s", CmdlineEscape().cmdline(cmd))
					subprocess.check_call(cmd, stdout = _log.subproc_target, stderr = _log.subproc_target)

				if self._args.capture_circuit or (self._args.write_back != _WriteBack.NoWriteback):
					# The circuit may have modified after it's settled. Retrieve it if user requests it.
					exported_circuit = await tx_rx({ "cmd": "circuit_export" })
					circuit.circuit_text = exported_circuit["data"]
					for change in self._args.simulation_change:
						if change == "reset_params":
							circuit.reset_presentation_parameters()
						else:
							raise NotImplementedError(change)

					if self._args.capture_circuit or (self._args.write_back == _WriteBack.ExternalFile):
						local_filename_circuit = circuit.automatic_filename
						output_filename_circuit = f"{self._args.output_dir}/{local_filename_circuit}"
						with open(output_filename_circuit, "w") as f:
							f.write(circuit.circuit_text)

					# Modify the presentation DOM
					if self._args.write_back == _WriteBack.Inline:
						self._dom_modified = circuit.modify_dom_source_inline() or self._dom_modified
					elif self._args.write_back == _WriteBack.ExternalFile:
						self._dom_modified = circuit.modify_dom_source_external_file("*") or self._dom_modified

		finally:
			if not self._args.no_shutdown:
				# Shutdown the simulator
				_log.debug("Shutting down the simulator")
				await tx_rx({ "cmd": "shutdown" }, expect_response = False)
		raise aiohttp.web_runner.GracefulExit()


	def _parse_circuits(self):
		circuits = [ ]
		(self._dom, presentation) = Presentation.parse_xml(self._args.infile)
		for slide in XMLTools.findall(presentation, "slide"):
			for circuit_node in XMLTools.findall(slide, "s:circuit"):
				circuit = CircuitJSCircuit.from_xml(circuit_node, find_file_function = self._lookup.lookup)
				name = circuit.get_presentation_parameter("name")
				if name is not None:
					if (len(self._args.only_circuit) == 0) or (name in self._args.only_circuit):
						circuits.append(circuit)
				else:
					_log.warning("Unnamed circuit will not be considered.")
		return circuits

	def run(self):
		self._lookup = FileLookup(self._args.include_dir)
		self._dom_modified = False
		self._connection = False
		self._msgid = 0
		self._circuits = self._parse_circuits()
		if len(self._circuits) == 0:
			_log.warning("No circuits found in presentation, nothing to do.")
			return 1

		with contextlib.suppress(FileExistsError):
			os.makedirs(self._args.output_dir)

		local_hostname = "127.0.0.1"
		local_uri = f"ws://{local_hostname}:{self._args.websocket_port}/ws"
		initial_config_args = {
			"ctz":	_empty_circuit_urlcomponent,
		}
		query = {
			"ws":			local_uri,
			"src":			f"circuitjs.html?{urllib.parse.urlencode(initial_config_args)}",
			"autoshutoff":	"1",
		}
		self._target_uri = f"{self._args.circuitjs_uri}?{urllib.parse.urlencode(query)}"
		if self._args.no_xdg_open:
			print(f"Point your browser to: {self._target_uri}")

		self._app = aiohttp.web.Application()
		self._app.on_startup.append(self._on_startup)

		self._app.router.add_route("GET", "/ws", self._websocket_handler)
		aiohttp.web.run_app(self._app, host = local_hostname, port = self._args.websocket_port)

		if self._dom_modified:
			rnd_filename = FileTools.base_random_file_on(self._args.infile)
			_log.debug("Writing modified presentation XML using tempfile %s", rnd_filename)
			with open(rnd_filename, "w") as f:
				self._dom.writexml(f)
			shutil.move(rnd_filename, self._args.infile)
		else:
			_log.debug("Presentation has not changed, not saving.")
		return 0



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import json
import base64
import zlib
import re
from .BaseAction import BaseAction
from .SpellcheckDictionary import SpellcheckDictionary, SpellcheckExceptionType

class ActionDictAdd(BaseAction):
	_EVIM_REGEX = re.compile(r"^(?P<base64>[A-Za-z0-9/+=]+)::(?P<filename>[^:]+)::(?P<line>\d+)::(?P<col>\d+)::(?P<msg>.*)$")

	def _choice(self, prompt, valid_answers, default_answer = None):
		while True:
			answer = input(prompt)
			if (answer == "") and (default_answer is not None):
				answer = default_answer
			if answer in valid_answers:
				return answer

	def _handle_current_finding(self):
		if self._current_finding < 0:
			self._current_finding = 0
		(match, data) = self._findings[self._current_finding]
		if self._dictionary.has_exception(offense = data["offense"], context = data["ctx"], rule_id = data["rule"]):
			# Already handeled the same finding in this session, do not handle it again
			self._current_finding += 1
			return

		print()
		print(f"Finding {self._current_finding + 1} of {len(self._findings)}:")
		print(match["msg"])
		print(f"Offense: > {data['offense']} <")
		print("   [A]dd word to dictionary")
		print("   Add word to [g]lobal dictionary (for all languages, e.g., names)")
		print("   Add specific [c]ontext to dictionary")
		print("   Jump to [p]previous finding")
		print("   Do [n]othing with this match (default)")
		answer = self._choice("Your choice: ", [ "a", "g", "c", "p", "n" ], default_answer = "n")
		if answer in [ "a", "g", "c" ]:
			exception_type = {
				"a": SpellcheckExceptionType.RuleWord,
				"g": SpellcheckExceptionType.GeneralWord,
				"c": SpellcheckExceptionType.RuleContext,
			}[answer]
			self._dictionary.add_exception(offense = data["offense"], context = data["ctx"], rule_id = data["rule"], exception_type = exception_type)
			self._current_finding += 1
		elif answer == "n":
			self._current_finding += 1
		elif answer == "p":
			self._current_finding -= 1
		else:
			raise NotImplementedError(answer)

	def _handle_findings(self):
		while self._current_finding < len(self._findings):
			self._handle_current_finding()

	def run(self):
		self._dictionary = SpellcheckDictionary.open_global_dict()
		self._current_finding = 0
		self._findings = [ ]

		try:
			with open(self._args.infile) as f:
				for line in f:
					line = line.rstrip("\r\n")
					match = self._EVIM_REGEX.fullmatch(line)
					if match is not None:
						match = match.groupdict()
						data = json.loads(zlib.decompress(base64.b64decode(match["base64"])))
						if not self._dictionary.has_exception(offense = data["offense"], context = data["ctx"], rule_id = data["rule"]):
							self._findings.append((match, data))
			self._handle_findings()
		finally:
			self._dictionary.write_if_modified()



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import collections
from pyradium.RendererCache import RendererCache
from pyradium.Exceptions import RendererRegistryException

class BaseRenderer():
	RendererResult = collections.namedtuple("RendererResult", [ "key", "data" ])
	_NAME = None
	_CACHE = True
	_RENDERER_CLASSES = { }
	_RENDERER_INSTANCES = { }

	@property
	def name(self):
		if self._NAME is None:
			raise NotImplementedError(__class__.__name__)
		return self._NAME

	@property
	def properties(self):
		return { "version": 0 }

	def rendering_key(self, property_dict):
		return None

	def render(self, property_dict):
		raise NotImplementedError(__class__.__name__)

	@classmethod
	def instanciate(cls, renderer_name, **kwargs):
		if renderer_name not in cls._RENDERER_INSTANCES:
			if renderer_name not in cls._RENDERER_CLASSES:
				raise RendererRegistryException(f"No renderer class registered for: {renderer_name}")
			renderer_class = cls._RENDERER_CLASSES[renderer_name]
			instance = renderer_class(**kwargs)
			if renderer_class._CACHE:
				instance = RendererCache(instance)
			cls._RENDERER_INSTANCES[renderer_name] = instance
		return cls._RENDERER_INSTANCES[renderer_name]

	@classmethod
	def register(cls, renderer_class):
		if renderer_class._NAME is None:
			raise RendererRegistryException("Class to be rendered needs to have a name.")
		if renderer_class._NAME in cls._RENDERER_CLASSES:
			raise RendererRegistryException(f"Duplicate renderer class name: {renderer_class._NAME}")
		cls._RENDERER_CLASSES[renderer_class._NAME] = renderer_class
		return renderer_class

if __name__ == "__main__":
	class DebugRenderer(BaseRenderer):
		@property
		def name(self):
			return "debug"

		@property
		def properties(self):
			return { "version": 1 }

		def render(self, property_dict):
			return {
				"text":		property_dict["letter"] * property_dict["count"],
				"bytes":	b"foobar" * 1000,
			}

	cache = RendererCache(DebugRenderer())
	print("Result = ", cache.render({
		"letter":	"Q",
		"count":	10,
	}))



#	pyradium - HTML presentation/slide show generator
#	Copyright (C) 2015-2022 Johannes Bauer
#
#	This file is part of pyradium.
#
#	pyradium is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	pyradium is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with pyradium; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import logging
from pyradium.Exceptions import UsageException
from pyradium.SVGTransformation import SVGTransformation

_log = logging.getLogger(__spec__.name)

class SVGAnimationMode(enum.Enum):
	Compose = "compose"				# Add new layers on top
	ComposeAll = "compose-all"		# Add new layers on top, but consider all layers from the start
	Replace = "replace"				# New layers replace lower layers

class SVGLayerTag(enum.Enum):
	NoStop = "nostop"				# Do not emit this layer
	Protect = "protect"				# Do not remove this layer on 'reset'
	Reset = "reset"					# Remove all but protected layers

class SVGAnimation():
	def __init__(self, svg_filename, animation_mode = None):
		if svg_filename is None:
			raise UsageException("SVG animation needs a source SVG filename, but none given.")
		if not svg_filename.lower().endswith(".svg"):
			raise UsageException(f"SVG animation only works with SVG input, but filename {svg_filename} does not end with '.svg'.")

		self._svg_filename = svg_filename
		if animation_mode is None:
			self._animation_mode = SVGAnimationMode.Compose
		else:
			try:
				self._animation_mode = SVGAnimationMode(animation_mode)
			except ValueError as e:
				raise UsageException(f"SVG animation mode '{animation_mode}' is  invalid (needs to be one of {SVGAnimationMode})") from e
		self._svg = SVGTransformation(svg_filename)

	def _determine_considered_layers(self) -> list[str]:
		if self._animation_mode == SVGAnimationMode.Compose:
			considered_layers = list(self._svg.visible_layer_ids)
		elif self._animation_mode in [ SVGAnimationMode.ComposeAll, SVGAnimationMode.Replace ]:
			considered_layers = list(self._svg.layer_ids)
		else:
			raise NotImplementedError(self._animation_mode)
		return considered_layers

	def _get_layer_tags(self, layer_id):
		tags = set()
		layer = self._svg.get_layer(layer_id)
		label = layer.label
		if ":" in label:
			(tags, _) = label.split(":", maxsplit = 1)
			for tag in tags:
				try:
					tag = SVGLayerTag(tag)
					tags.add(tag)
				except ValueError as e:
					_log.warning("Unknown layer tag in %s layer %s: %s", self._svg_filename, layer_id, tag)
		return tags

	def _show_layer(self, layer_id):
		self._svg_transforms.append({
			"cmd":			"show_layer",
			"layer_id":		layer_id,
		})
		if SVGLayerTag.Protect not in self._layer_tags[layer_id]:
			self._shown_unprotected_layers.add(layer_id)

	def _hide_layer(self, layer_id):
		self._svg_transforms.append({
			"cmd":			"hide_layer",
			"layer_id":		layer_id,
		})
		if layer_id in self._shown_unprotected_layers:
			self._shown_unprotected_layers.remove(layer_id)

	def _hide_all_layers(self):
		for layer_id in self._considered_layers:
			self._hide_layer(layer_id)

	def render(self):
		# Determine which layers we consider for the anymation in the first place
		self._considered_layers = self._determine_considered_layers()

		# Then get all the tags associated with those layers
		self._layer_tags = { layer_id: self._get_layer_tags(layer_id) for layer_id in self._considered_layers }

		# Find out which layers are protected
		self._protected_layer_ids = set(layer_id for (layer_id, layer_tags) in self._layer_tags.items() if SVGLayerTag.Protect in layer_tags)

		# Store commands to hide all layers first
		self._svg_transforms = [ ]
		self._shown_unprotected_layers = set()
		self._hide_all_layers()

		# Then go through the layers one-by-one and render then as appropriate
		previous_layer_id = None
		for layer_id in self._considered_layers:
			tags = self._layer_tags[layer_id]

			# Determine if slide is causing a reset, i.e. removal of below
			# layers which are unprotected
			if SVGLayerTag.Reset in tags:
				# Hide all below layers but those which are protected
				for below_layer_id in self._shown_unprotected_layers:
					self._hide_layer(below_layer_id)

			# Then show the current layer
			self._show_layer(layer_id)

			# If we're in a replacing mode, then hide the previous layer
			if (self._animation_mode == SVGAnimationMode.Replace) and (previous_layer_id is not None):
				self._hide_layer(previous_layer_id)

			# Emit it as a frame if it's not marked as "nostop"
			if SVGLayerTag.NoStop not in tags:
				yield self._svg_transforms
			previous_layer_id = layer_id
